
//-------------------------------------------------------------------------------//
// Author: Stefan Lörwald, Universität Heidelberg                                //
// License: CC BY-NC 4.0 http://creativecommons.org/licenses/by-nc/4.0/legalcode //
//-------------------------------------------------------------------------------//

#define COMPILE_TEMPLATE_METHOD_ADJACENCY_DECOMPOSITION_IMPLEMENTATION
#include "method_adjacency_decomposition_implementation.h"
#undef COMPILE_TEMPLATE_METHOD_ADJACENCY_DECOMPOSITION_IMPLEMENTATION

#include <algorithm>
#include <future>
#include <iostream>
#include <list>

#include "algorithm_facet_classes.h"
#include "algorithm_fourier_motzkin_elimination.h"
#include "algorithm_map_operations.h"
#include "algorithm_matrix_operations.h"
#include "algorithm_rotation.h"
#include "algorithm_row_operations.h"
#include "joining_thread.h"

namespace
{
   template <typename Integer>
   std::future<void> initializePool(JobManager<Integer>&, const ConvexHull<Integer>&, const Names&, Maps&, const Inequalities<Integer>& known_facets, Equations<Integer>&);

   template <typename Integer>
   std::future<void> initializePool(JobManagerProxy<Integer>&, const ConvexHull<Integer>&, const Names&, Maps&, const Inequalities<Integer>& known_facets, Equations<Integer>&);
}

template <template <typename> class JobManagerType, typename Integer>
void implementation::adjacencyDecomposition(const std::tuple<ConvexHull<Integer>, Names, Maps>& data, const Inequalities<Integer>& known_facets, const int thread_count, const int node_count)
{
   const ConvexHull<Integer>& vertices = std::get<0>(data);
   const Names& names = std::get<1>(data);
   Maps maps = std::get<2>(data);
   JobManagerType<Integer> job_manager(names, node_count, thread_count);
   Equations<Integer> equations;
   std::list<JoiningThread> threads;
   auto future = initializePool(job_manager, vertices, names, maps, known_facets, equations);
   for ( int i(0); i < thread_count; ++i )
   {
      threads.emplace_front([&]()
      {
         while ( true )
         {
            const auto facet = job_manager.get();
            if ( facet.empty() )
            {
               break;
            }
            const auto facets = algorithm::rotation(vertices, facet, maps);
            job_manager.put(facets);
         }
      });
   }
   future.wait();
}

namespace
{
   template <typename Integer>
   std::future<void> initializePool(JobManager<Integer>& manager, const ConvexHull<Integer>& conv, const Names& names, Maps& maps, const Inequalities<Integer>& known_facets, Equations<Integer>& equations)
   {
      equations = algorithm::extractEquations(conv);
      if ( !equations.empty() )
      {
         std::cout << "Equations:\n";
         prettyPrint(std::cout, equations, names, "=");
         std::cout << "\n";
         maps = algorithm::normalize(maps, equations);
      }
      std::cout << "Inequalities:\n";
      // Initialize the process (so that other processes start).
      if ( !known_facets.empty() )
      {
         auto tmp = normalize(known_facets.back(), equations);
         tmp = algorithm::classRepresentative(tmp, maps);
         manager.put(Matrix<Integer>{tmp});
      }
      else
      {
         auto facets = algorithm::fourierMotzkinEliminationHeuristic(conv);
         for ( auto& facet : facets )
         {
            facet = algorithm::classRepresentative(facet, maps);
         }
         manager.put(facets);
      }
      // Add the remaining known facets from file asynchronously.
      auto future = std::async(std::launch::async, [&]()
      {
         for ( const auto& facet : known_facets )
         {
            auto tmp = normalize(facet, equations);
            tmp = algorithm::classRepresentative(tmp, maps);
            manager.put(tmp);
         }
      });
      return future;
   }

   template <typename Integer>
   std::future<void> initializePool(JobManagerProxy<Integer>&, const ConvexHull<Integer>& conv, const Names&, Maps& maps, const Inequalities<Integer>&, Equations<Integer>& equations)
   {
      equations = algorithm::extractEquations(conv);
      if ( !equations.empty() )
      {
         maps = algorithm::normalize(maps, equations);
      }
      // only the manager on the root node performs a heuristic to get initial facets.
      auto future = std::async(std::launch::async, [](){});
      return future;
   }
}

